ALL_OF = "all_of"
DAYS = "days"
ENDS_EARLIER_THAN = "ends_earlier_than"
ENDS_LATER_THAN = "ends_later_than"
NONE_OF = "none_of"
ONE_OF = "one_of"
STARTS_EARLIER_THAN = "starts_earlier_than"
STARTS_LATER_THAN = "starts_later_than"
WEEKDAY = "weekday"
WEEKEND = "weekend"
