# When updating the version, also update it in the docs.
__version__ = "0.4.5"

# The following are set by Travis CI at build time.
__build_number__ = None
__build_timestamp__ = None
__build_href__ = None
__git_commit_hash__ = None

# NOTE: the following were set by Travis CI
__version__ = "0.4.5"
__build_number__ = 850
__build_timestamp__ = 1589467080
__build_href__ = "https://travis-ci.org/jamespfennell/transiter/jobs/687048115"
__git_commit_hash__ = "74e61ad1df2e2d56308128158ac399c444be9d81"
